/* Copyright (c) 2011, S V N Vishwanathan
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Authors: S V N Vishwanathan (vishy@stat.purdue.edu)
 *          Xinhua Zhang (xinhua.zhang.cs@gmail.com)
 *
*/

#include "tao.h"
#include <cmath>
#include <cassert>
#include <algorithm>

#include "appctx.hpp"
#include "loadfile.hpp"

#include "hingeloss.hpp"
#include "rocscoreloss.hpp"
#include "prbeploss.hpp"

static  char help[]="soft hinge loss and its generalizations using TAO \n\
Input parameters are:\n                                                 \
  -algo    <int>      : 1=binary hinge loss (default) 2=ROCArea loss 3=PRBEP\n  \
  -data    <filename> : file which contains data in Petsc binary format\n \
  -labels  <filename> : file which contains labels in Petsc binary format\n \
  -tdata   <filename> : file which contains test data in Petsc binary format\n \
  -tlabels <filename> : file which contains test labels in Petsc binary format\n \
  -lambda  <double>   : regularization parameter (default 1e-6)\n       \
  -eps     <double>   : smoothing parameter (default 0.1)\n \
  -buf     <bool>     : use buffered line search (default 1)\n \
  -libsvm  <bool>     : use LibSVM format feature/label file (default 0)\n \n";


// Function pointers to loss and gradient computation functions 
PetscErrorCode (*loss_grad)(PetscScalar *fx_array, 
                            PetscScalar *labels_array, 
                            PetscScalar *grad_array, 
                            PetscInt    n, 
                            PetscScalar *loss,
                            AppCtx      *user);    

PetscErrorCode (*orig_loss)(PetscScalar *fx_array, 
                            PetscScalar *labels_array, 
                            PetscInt    n, 
                            PetscScalar *loss,
                            AppCtx      *user);

// Initialize and finalize the loss 
PetscErrorCode (*init)(AppCtx *user);

PetscErrorCode (*finalize)(void);

enum { HINGELOSS=1, ROCAREA=2, PRBEP=3 };


// Forwarding function
PetscErrorCode ScalarLossGradBForce(TaoSolver tao, Vec w, double *f, Vec G, void *ctx);

// Predict labels on test data
PetscErrorCode Predict(Vec w, AppCtx* user);

// Check Generalization of intermediate solutions 
PetscErrorCode CheckGeneralization(TaoSolver tao, void *ctx);

// Check Generalization of intermediate solutions 
PetscErrorCode PrintFunVal(TaoSolver tao, void *ctx);

// Read in the feature and label data
PetscErrorCode LoadData(AppCtx *user);

// Parse the arguments in the command line
PetscErrorCode LoadOptions(AppCtx *user);

// Print a summary of the settings
PetscErrorCode PrintSetting(AppCtx *user);

// Allocate the space for work
PetscErrorCode AllocateWorkSpace(Vec *w, AppCtx *user);

// Destroy the work space
PetscErrorCode DestroyWorkSpace(Vec *w, AppCtx *user);

// At the end of a round of line search, compute the gradient
PetscErrorCode FinalGradLineSearch(TaoLineSearch ls, Vec w, Vec g, void *ctx);

// Return the function value and grad*ls_direction
PetscErrorCode LineSearchObjAndGTS(TaoLineSearch ls, Vec w, Vec dir, PetscReal *f, PetscReal *gts, void *ctx);


#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc, char **argv)
{
  AppCtx            user;                  
  char              hostname[1024];
  PetscErrorCode    info;              
  Vec               w;                     // solution vector
  TaoSolver         tao;                   
  TaoSolverTerminationReason reason;
  
  // Initialize TAO and PETSc
  PetscInitialize(&argc, &argv, (char *)0, help);
  TaoInitialize(&argc, &argv, (char *)0, help);
  
  MPI_Comm_rank(PETSC_COMM_WORLD, &user.rank);
  MPI_Comm_size(PETSC_COMM_WORLD, &user.size);
  PetscPrintf(PETSC_COMM_WORLD, "%d machines found\n", user.size);
  gethostname(hostname, 1024);
  PetscSynchronizedPrintf(PETSC_COMM_WORLD, "%s\n", hostname); 
  PetscSynchronizedFlush(PETSC_COMM_WORLD);


  // Check for command line arguments to override defaults
  LoadOptions(&user);
  LoadData(&user);
  AllocateWorkSpace(&w, &user);   // Allocate space 
  
  if (user.rank==0){
    PetscScalar *labels_array;
    info=VecGetArray(user.labelsloc,&labels_array); CHKERRQ(info);
    
    // Paranoia 
    for (PetscInt i=0; i<user.m; i++)
      assert(fabs(labels_array[i]+1)<1e-5 || fabs(labels_array[i]-1)<1e-5);
    
    user.m_pos = user.m_neg = 0;
    for (PetscInt i=0; i<user.m; i++)
      if (labels_array[i]>0) user.m_pos++;
      else user.m_neg++;
    
    info = VecRestoreArray(user.labelsloc,&labels_array); CHKERRQ(info);    
  }
  
  if (user.rank==0) {
    init(&user);
    PrintSetting(&user);
  }


  ///////////////////////////////////////////////////////////////////
  // TAO code begins here
  
  // Create TAO solver with desired solution method 
  info = TaoCreate(PETSC_COMM_WORLD, &tao); CHKERRQ(info);
  info = TaoSetType(tao, "tao_lmvm"); CHKERRQ(info);

  info = TaoSetInitialVector(tao, w); CHKERRQ(info);
  info = TaoSetObjectiveAndGradientRoutine(tao, ScalarLossGradBForce, &user); CHKERRQ(info);

  // Set routines for function, gradient, Hessian evaluation
  if (user.use_buf_ls)  {    
    user.new_round_ls = PETSC_TRUE;

    TaoLineSearch ls;
    info = TaoGetLineSearch(tao, &ls); CHKERRQ(info);

    // Specify how to compute the objective value and the inner product of gradient and line search direction
    // Should only use the step size based on buffered inner products
    info = TaoLineSearchSetObjectiveAndGTSRoutine(ls, LineSearchObjAndGTS, &user); CHKERRQ(info);

    // Specify how to compute the gradient upon completion of one round of line search
    info = TaoLineSearchSetGradientRoutine(ls, FinalGradLineSearch, &user); CHKERRQ(info);
  }
  else  {
  }
  
  user.num_fn_eval = 0;
  info = TaoSetTolerances(tao,1e-10,0,1e-5,0,0); CHKERRQ(info);

  // Check for TAO command line options 
  info = TaoSetFromOptions(tao); CHKERRQ(info);

  // Print intermediate function values 
  info = TaoSetMonitor(tao, PrintFunVal, &user, NULL); CHKERRQ(info);
    
  // Start the timer 
  if (user.rank==0)
    user.opt_timer.start();  

  // Solve the Application   
  info = TaoSolve(tao); CHKERRQ(info);
 
  // Stop the timer
  if (user.rank==0)
    user.opt_timer.stop();
  
  // Print statistics 
  if (user.rank==0){
    PetscPrintf(PETSC_COMM_WORLD, "Total CPU time: %g seconds \n", user.opt_timer.cpu_total);
    PetscPrintf(PETSC_COMM_WORLD, "Total wallclock time: %g seconds \n", user.opt_timer.wallclock_total);
    PetscPrintf(PETSC_COMM_WORLD, "Obj grad CPU time: %g seconds \n", user.objgrad_timer.cpu_total);
    PetscPrintf(PETSC_COMM_WORLD, "Obj grad wallclock time: %g seconds \n", user.objgrad_timer.wallclock_total);
  }
  
  Predict(w, &user);

  // Free some space
  DestroyWorkSpace(&w, &user);

  // Get termination information
  info=TaoGetTerminationReason(tao,&reason); CHKERRQ(info);
  if (reason <= 0)
    PetscPrintf(PETSC_COMM_WORLD, "TAO convergence failure with reason %D \n", reason);
  else{
    PetscPrintf(PETSC_COMM_WORLD, "Successful convergence with reason %D \n", reason);
    // VecView(w, PETSC_VIEWER_STDOUT_WORLD);    
  }

  // Free TAO data structures 
  info = TaoDestroy(&tao); CHKERRQ(info);
  
  if (user.rank==0)
    finalize();
  
  // Finalize TAO and PETSc
  info = TaoFinalize(); CHKERRQ(info);
  info = PetscFinalize(); CHKERRQ(info);
  
  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "LoadOptions"
PetscErrorCode LoadOptions(AppCtx *user)
{
  PetscBool         flg;
  PetscErrorCode    info;

  user->lambda = 1e-6; 
  user->algo = HINGELOSS; 
  user->eps = 0.1; 
  user->use_buf_ls = PETSC_TRUE; 
  user->libsvm = PETSC_FALSE; 

  info = PetscOptionsGetReal(PETSC_NULL, "-lambda", &user->lambda, &flg); CHKERRQ(info); 
  info = PetscOptionsGetInt(PETSC_NULL, "-algo", &user->algo, &flg); CHKERRQ(info); 
  info = PetscOptionsGetReal(PETSC_NULL, "-eps", &user->eps, &flg); CHKERRQ(info); 
  info = PetscOptionsGetBool(PETSC_NULL, "-buf", &user->use_buf_ls, &flg); CHKERRQ(info); 
  assert(user->lambda > 0.0); 
  assert(user->eps > 0.0); 

  switch(user->algo){
  case ROCAREA:
    loss_grad = rocscore::loss_grad; 
    orig_loss = rocscore::orig_loss; 
    init = rocscore::init; 
    finalize = rocscore::finalize; 
    strcpy(user->algo_string, "ROCArea"); 
    break; 
  case PRBEP:
    loss_grad = prbep::loss_grad; 
    orig_loss = prbep::orig_loss; 
    init = prbep::init; 
    finalize = prbep::finalize; 
    strcpy(user->algo_string, "PRBEP"); 
    break; 
  case HINGELOSS:
  default:
    loss_grad = hingeloss::loss_grad; 
    orig_loss = hingeloss::orig_loss; 
    init = hingeloss::init; 
    finalize = hingeloss::finalize; 
    strcpy(user->algo_string, "Hinge"); 
    break; 
  }

  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "LoadData"
PetscErrorCode LoadData(AppCtx *user_main)
{
  PetscViewer        view;
  PetscErrorCode     info;
  PetscBool          flg;

  info=PetscOptionsGetBool(PETSC_NULL, "-libsvm", &user_main->libsvm, &flg); CHKERRQ(info);

  if (!user_main->libsvm)
  {
    PetscBool         data_flg, labels_flg;

    info = PetscOptionsGetString(PETSC_NULL, "-data", user_main->data_path, PETSC_MAX_PATH_LEN-1, &data_flg); CHKERRQ(info);
    info = PetscOptionsGetString(PETSC_NULL, "-labels", user_main->labels_path, PETSC_MAX_PATH_LEN-1, &labels_flg); CHKERRQ(info);

    if (!data_flg || !labels_flg)
      SETERRQ(PETSC_ERR_SUP, 1, "No train files specified!");

    info = PetscOptionsGetString(PETSC_NULL, "-tdata", user_main->tdata_path, PETSC_MAX_PATH_LEN-1, &data_flg); CHKERRQ(info);
    info = PetscOptionsGetString(PETSC_NULL, "-tlabels", user_main->tlabels_path, PETSC_MAX_PATH_LEN-1, &labels_flg); CHKERRQ(info);

    if (!data_flg || !labels_flg)
      SETERRQ(PETSC_ERR_SUP, 1, "No test files specified!");

    info = PetscViewerBinaryOpen(PETSC_COMM_WORLD, user_main->data_path, FILE_MODE_READ, &view); CHKERRQ(info);
    info = MatCreate(PETSC_COMM_WORLD, &user_main->data); CHKERRQ(info);  
    info = MatSetType(user_main->data, MATSEQAIJ); CHKERRQ(info);  
    info = MatLoad(user_main->data, view); CHKERRQ(info);   // , MATAIJ, 
    info = PetscViewerDestroy(&view); CHKERRQ(info);

    info = PetscViewerBinaryOpen(PETSC_COMM_WORLD, user_main->labels_path, FILE_MODE_READ, &view); CHKERRQ(info);
    info = VecCreate(PETSC_COMM_WORLD, &user_main->labels); CHKERRQ(info);
    info = VecLoad(user_main->labels, view); CHKERRQ(info);
    info = PetscViewerDestroy(&view); CHKERRQ(info);

    info = MatGetSize(user_main->data, &user_main->m, &user_main->dim); CHKERRQ(info);

    info = PetscViewerBinaryOpen(PETSC_COMM_WORLD, user_main->tdata_path, FILE_MODE_READ, &view); CHKERRQ(info);
    info = MatCreate(PETSC_COMM_WORLD, &user_main->tdata); CHKERRQ(info);  
    info = MatSetType(user_main->tdata, MATAIJ); CHKERRQ(info);  
    info = MatLoad(user_main->tdata, view); CHKERRQ(info);
    info = PetscViewerDestroy(&view); CHKERRQ(info);

    info = PetscViewerBinaryOpen(PETSC_COMM_WORLD, user_main->tlabels_path, FILE_MODE_READ, &view); CHKERRQ(info);
    info = VecCreate(PETSC_COMM_WORLD, &user_main->tlabels); CHKERRQ(info);    
    info = VecLoad(user_main->tlabels, view); CHKERRQ(info);
    info = PetscViewerDestroy(&view); CHKERRQ(info);

    info = MatGetSize(user_main->tdata, &user_main->tm, &user_main->tdim); CHKERRQ(info);
    assert(user_main->tdim == user_main->dim);
  }
  else
  {
    // Always add bias
    AppCtx_file    fuser; 
    AppCtx_file    tfuser; 
    FILE           *data_file;
    FILE           *tdata_file;

    info=PetscOptionsGetString(PETSC_NULL, "-train", user_main->data_path, PETSC_MAX_PATH_LEN-1, &flg); CHKERRQ(info);
    if (!flg)
      SETERRQ(PETSC_ERR_SUP, 1, "No training data files specified!");
    info=PetscOptionsGetString(PETSC_NULL, "-test", user_main->tdata_path, PETSC_MAX_PATH_LEN-1, &flg); CHKERRQ(info);
    if (!flg)
      SETERRQ(PETSC_ERR_SUP, 1, "No test data files specified!");

    fuser.data = &user_main->data;
    fuser.labels = &user_main->labels;
    tfuser.data = &user_main->tdata;
    tfuser.labels = &user_main->tlabels;

    PetscPrintf(PETSC_COMM_WORLD, "Start parsing the training data\n");
    fuser.flg = tfuser.flg = PETSC_TRUE;
    info=PetscFOpen(PETSC_COMM_SELF, user_main->data_path,"r",&data_file); CHKERRQ(info); 
    fuser.nnz_array=NULL;
    if (user_main->rank==0){
      parse_file(data_file,&fuser);    
      rewind(data_file);   // Set file pointer to beginning of file
    }

    PetscPrintf(PETSC_COMM_WORLD, "Start parsing the test data\n");
    tfuser.nnz_array=NULL;
    info=PetscFOpen(PETSC_COMM_SELF, user_main->tdata_path,"r",&tdata_file); CHKERRQ(info); 
    if (user_main->rank==0){
      parse_file(tdata_file,&tfuser);    
      rewind(tdata_file);   // Set file pointer to beginning of file
      // Choose the larger of the two dimensions
      if (fuser.dim > tfuser.dim)
        tfuser.dim=fuser.dim;
      else
        fuser.dim=tfuser.dim;
    }

    MPI_Bcast(&tfuser.dim,1,MPIU_INT,0,PETSC_COMM_WORLD);  
    MPI_Bcast(&tfuser.m,1,MPIU_INT,0,PETSC_COMM_WORLD);
    MPI_Bcast(&tfuser.maxnnz,1,MPIU_INT,0,PETSC_COMM_WORLD);  
    MPI_Bcast(&tfuser.maxlen,1,MPIU_INT,0,PETSC_COMM_WORLD);
    if (user_main->rank!=0){
      info=PetscMalloc(tfuser.m*sizeof(PetscInt),&tfuser.nnz_array); CHKERRQ(info);
    }
    MPI_Bcast(tfuser.nnz_array,tfuser.m,MPIU_INT,0,PETSC_COMM_WORLD);

    MPI_Bcast(&fuser.dim,1,MPIU_INT,0,PETSC_COMM_WORLD);  
    MPI_Bcast(&fuser.m,1,MPIU_INT,0,PETSC_COMM_WORLD);
    MPI_Bcast(&fuser.maxnnz,1,MPIU_INT,0,PETSC_COMM_WORLD);  
    MPI_Bcast(&fuser.maxlen,1,MPIU_INT,0,PETSC_COMM_WORLD);
    if (user_main->rank!=0){
      info=PetscMalloc(fuser.m*sizeof(PetscInt),&fuser.nnz_array); CHKERRQ(info);
    }
    MPI_Bcast(fuser.nnz_array,fuser.m,MPIU_INT,0,PETSC_COMM_WORLD);

    PetscPrintf(PETSC_COMM_WORLD,"Start loading the training data\n");
    if (user_main->size==1)
      fill_arrays(data_file,&fuser);
    else
      fill_arrays_parallel(data_file,&fuser);
    info=PetscFree(fuser.nnz_array); CHKERRQ(info);
    PetscFClose(PETSC_COMM_SELF,data_file); 

    PetscPrintf(PETSC_COMM_WORLD,"Start loading the test data\n");
    if (user_main->size==1)
      fill_arrays(tdata_file,&tfuser);
    else{
      fill_arrays_parallel(tdata_file,&tfuser);
    }
    info=PetscFree(tfuser.nnz_array); CHKERRQ(info);
    PetscFClose(PETSC_COMM_SELF,tdata_file); 

    info=MatGetSize(user_main->data,&user_main->m,&user_main->dim); CHKERRQ(info);
    info=MatGetSize(user_main->tdata,&user_main->tm,&user_main->tdim); CHKERRQ(info);
  }
  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "PrintSetting"
PetscErrorCode PrintSetting(AppCtx *user)
{
  PetscErrorCode info;

  info = PetscPrintf(PETSC_COMM_SELF, "Number of processors: %D\n", user->size); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "lambda: %e, eps: %e\n", user->lambda, user->eps); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "algo: %s\n", user->algo_string); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "Buffered line search: %d\n", user->use_buf_ls); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "Input in LibSVM format: %d\n", user->libsvm); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "train: %s\n", user->data_path); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "m_pos: %D, m_neg = %D\n", user->m_pos, user->m_neg); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "dim: %D\n", user->dim);


  info = PetscPrintf(PETSC_COMM_SELF, "test: %s\n", user->tdata_path); CHKERRQ(info);
  info = PetscPrintf(PETSC_COMM_SELF, "m_test: %D\n", user->tm); CHKERRQ(info);
  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "AllocateWorkSpace"
PetscErrorCode AllocateWorkSpace(Vec *w, AppCtx *user)
{
  PetscErrorCode     info;              

  info = VecCreate(PETSC_COMM_WORLD,w); CHKERRQ(info); 
  info = VecSetSizes(*w, PETSC_DECIDE, user->dim); CHKERRQ(info); 
  info = VecSetFromOptions(*w); CHKERRQ(info); 
  info = PetscObjectSetName((PetscObject) *w, "Solution"); CHKERRQ(info); 
  info = VecSet(*w,0.0); CHKERRQ(info); 
  info = VecAssemblyBegin(*w); CHKERRQ(info); 
  info = VecAssemblyEnd(*w); CHKERRQ(info); 

  info = VecDuplicate(user->labels, &user->fx); CHKERRQ(info); 
  info = PetscObjectSetName((PetscObject) user->fx,"Function Values"); CHKERRQ(info); 
  info = VecAssemblyBegin(user->fx); CHKERRQ(info); 
  info = VecAssemblyEnd(user->fx); CHKERRQ(info); 

  info = VecDuplicate(user->fx, &user->grad); CHKERRQ(info); 
  info = PetscObjectSetName((PetscObject) user->grad, "Gradient Coefficients"); CHKERRQ(info); 
  info = VecAssemblyBegin(user->grad); CHKERRQ(info); 
  info = VecAssemblyEnd(user->grad); CHKERRQ(info); 

  info = VecScatterCreateToZero(user->fx,&user->scatter,&user->fxloc); CHKERRQ(info); 
  info = VecAssemblyBegin(user->fxloc); CHKERRQ(info); 
  info = VecAssemblyEnd(user->fxloc); CHKERRQ(info); 

  info = VecDuplicate(user->fxloc, &user->gradloc); CHKERRQ(info); 
  info = VecAssemblyBegin(user->gradloc); CHKERRQ(info); 
  info = VecAssemblyEnd(user->gradloc); CHKERRQ(info); 

  info = VecDuplicate(user->fxloc, &user->labelsloc); CHKERRQ(info); 
  info = VecAssemblyBegin(user->labelsloc); CHKERRQ(info); 
  info = VecAssemblyEnd(user->labelsloc); CHKERRQ(info); 

  info = VecScatterBegin(user->scatter,user->labels,user->labelsloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info); 
  info = VecScatterEnd(user->scatter,user->labels,user->labelsloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info); 

  if (user->use_buf_ls)  {
    info = VecDuplicate(user->fx, &user->XMultW); CHKERRQ(info);
    info = PetscObjectSetName((PetscObject) user->XMultW, "XMultW"); CHKERRQ(info);

    info = VecDuplicate(user->fx, &user->XMultDir); CHKERRQ(info);
    info = PetscObjectSetName((PetscObject) user->XMultDir, "XMultDir"); CHKERRQ(info);
  }

  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "DestroyWorkSpace"
PetscErrorCode DestroyWorkSpace(Vec *w, AppCtx *user)
{
  PetscErrorCode     info;              

  // Free some space
  info = MatDestroy(&user->data); CHKERRQ(info);
  info = VecDestroy(&user->labels); CHKERRQ(info);
  info = VecDestroy(&user->fx); CHKERRQ(info);
  info = VecDestroy(&user->grad); CHKERRQ(info);

  info = VecDestroy(&user->fxloc); CHKERRQ(info);
  info = VecDestroy(&user->labelsloc); CHKERRQ(info);
  info = VecDestroy(&user->gradloc); CHKERRQ(info);

  info = VecScatterDestroy(&user->scatter); CHKERRQ(info);

  info = MatDestroy(&user->tdata); CHKERRQ(info);
  info = VecDestroy(&user->tlabels); CHKERRQ(info);
  info = VecDestroy(w); CHKERRQ(info);

  if (user->use_buf_ls)
  {
    info = VecDestroy(&user->XMultW); CHKERRQ(info);
    info = VecDestroy(&user->XMultDir); CHKERRQ(info);
  }

  return 0;
}
 

#undef __FUNCT__
#define __FUNCT__ "PrintFunVal"
// PrintFunVal - Print intermediate function values 
 
// Input Parameters:
// tao     - tao solver
// ctx     - user defined application context
 
// Output Parameters:
// None. Display intermediate function values  
 PetscErrorCode PrintFunVal(TaoSolver tao, void *ctx){
   AppCtx *user = (AppCtx *) ctx;
   PetscErrorCode info;
   PetscScalar f, ori_f;
   PetscScalar gnorm;
   PetscInt iter;
   PetscScalar reg=0.0;
   PetscScalar *fx_array, *labels_array;
   Vec w;

   if (user->rank==0)
     user->opt_timer.stop();

   info = TaoGetSolutionVector(tao,&w); CHKERRQ(info);

   info = VecDot(w,w,&reg); CHKERRQ(info);
   reg *= user->lambda/2.0;

   info = MatMult(user->data,w,user->fx); CHKERRQ(info);

   // All nodes send fx to the master node
   info = VecScatterBegin(user->scatter,user->fx,user->fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);
   info = VecScatterEnd(user->scatter,user->fx,user->fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);

   // Master node computes the function and gradient coefficients 
   double partloss = 0.0;
   if (user->rank == 0){
     info = VecGetArray(user->fxloc,&fx_array); CHKERRQ(info);
     info = VecGetArray(user->labelsloc,&labels_array); CHKERRQ(info);

     orig_loss(fx_array,labels_array,user->m,&partloss,user);

     info = VecRestoreArray(user->fxloc,&fx_array); CHKERRQ(info);
     info = VecRestoreArray(user->labelsloc,&labels_array); CHKERRQ(info);
   }

   MPI_Barrier(PETSC_COMM_WORLD);
   ori_f = 0.0;
   MPI_Allreduce(&partloss,&ori_f,1,MPI_DOUBLE,MPI_SUM,PETSC_COMM_WORLD);
   ori_f += reg;

   info = TaoGetSolutionStatus(tao, &iter, &f, &gnorm, NULL, NULL, NULL);
   PetscPrintf(PETSC_COMM_WORLD, "Iter: %d, ori_f: %g, f: %g, gnorm: %g \n", iter,ori_f,f,gnorm);

   PetscPrintf(PETSC_COMM_WORLD,"num_fn_eval %d CPU: %g Wallclock: %g grad_t: %g matvec_t %g\n", 
                user->num_fn_eval, user->opt_timer.cpu_total, user->opt_timer.wallclock_total,
                user->objgrad_timer.cpu_total, user->matvec_timer.cpu_total);

   // Check generalization performance
   info = Predict(w, user); CHKERRQ(info);

   if (user->rank==0)
     user->opt_timer.start();
   return 0;
 }
 
 
#undef __FUNCT__
#define __FUNCT__ "Predict"
// Predict - Predict on test set and evaluate accuracy

// Input Parameters:
// w       - current weight vector
// user    - Application context

// Output Parameters:
// None. Display accuracy 

PetscErrorCode Predict(Vec w, AppCtx* user){
  PetscErrorCode info;
  PetscScalar *fx_array,*labels_array;
  Vec fx,fxloc,labelsloc;
  VecScatter  scatter;   
  
  info = VecCreate(PETSC_COMM_WORLD,&fx); CHKERRQ(info);
  info = VecSetFromOptions(fx); CHKERRQ(info);
  info = VecSetSizes(fx,PETSC_DECIDE,user->tm); CHKERRQ(info);
  info = PetscObjectSetName((PetscObject) fx,"Prediction on Test Set"); CHKERRQ(info);
  
  info = MatMult(user->tdata,w,fx); CHKERRQ(info);
  
  info = VecScatterCreateToZero(fx,&scatter,&fxloc); CHKERRQ(info);
  info = VecAssemblyBegin(fxloc); CHKERRQ(info);
  info = VecAssemblyEnd(fxloc); CHKERRQ(info);

  info = VecDuplicate(fxloc, &labelsloc); CHKERRQ(info);
  info = VecAssemblyBegin(labelsloc); CHKERRQ(info);
  info = VecAssemblyEnd(labelsloc); CHKERRQ(info);

  // gather the distributed fx into the whole fxloc
  info = VecScatterBegin(scatter,fx,fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);
  info = VecScatterEnd(scatter,fx,fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);
  
  // gather the distributed tlabels into the whole labelsloc
  info = VecScatterBegin(scatter,user->tlabels,labelsloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);
  info = VecScatterEnd(scatter,user->tlabels,labelsloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);

  info = VecScatterDestroy(&scatter); CHKERRQ(info);
  info = VecGetArray(fxloc,&fx_array); CHKERRQ(info);
  info = VecGetArray(labelsloc,&labels_array); CHKERRQ(info);

  if (user->rank==0)
  {
    if (user->algo==1){
      PetscInt correct=0;

      // Compute predictions
      for (PetscInt i=0; i<user->tm; i++){
        if (fx_array[i]*labels_array[i]>0)
          correct++;
      }

      printf("Test: %ld Correct: %ld Accuracy: %f \n", 
              (long int)user->tm, (long int)correct, (100.0*correct)/user->tm);

    } else if (user->algo==2 || user->algo==3){

      PetscInt TP=0;    // true positive
      PetscInt TN=0;    // true negative
      PetscInt FP=0;    // false positive
      PetscInt FN=0;    // false negative

      for (PetscInt i=0; i<user->tm; i++){
        if (fx_array[i]>0  && labels_array[i]>0)   TP++;
        if (fx_array[i]>0  && labels_array[i]<=0)  FP++;
        if (fx_array[i]<=0 && labels_array[i]>0)   FN++;
        if (fx_array[i]<=0 && labels_array[i]<=0)  TN++;
      }

      PetscInt* idx=0;
      info=PetscMalloc(user->tm*sizeof(PetscInt),&idx); CHKERRQ(info);
      for (PetscInt i=0; i<user->tm; i++)
        idx[i]=i;

      indirect_less_than<PetscScalar> ilt_fx(fx_array);    
      std::sort(idx, idx+user->tm, ilt_fx);

      // count swapped pairs i.e., misclassified points
      PetscScalar sum = 0.0;
      PetscInt num_fp=FP+TN,num_fn=0,cross_pt=0,n_pos=0,n_neg=0;

      for (PetscInt i=0; i<user->tm; i++){
        if (labels_array[idx[i]]<0){
          sum+=n_pos;
          n_neg++;
          num_fp--;
        } else {
          n_pos++;
          num_fn++;
        }
        if (num_fp==num_fn) 
          cross_pt=num_fp;
      }
      assert(n_neg==FP+TN && n_pos==TP+FN);

      double accuracy = (TP+TN)*100.0/user->tm;
      double roc = 100.0*(1.0 - ((sum/n_neg)/n_pos));
      double prbep = (n_pos - cross_pt) * 100.0 / n_pos;

      printf("Accuracy: %f ROCArea: %f PRBEP: %f\n", accuracy, roc, prbep);

      info=PetscFree(idx); CHKERRQ(info);
    }
  }
  
  // Restore vectors 
  info = VecRestoreArray(fxloc,&fx_array); CHKERRQ(info);
  info = VecRestoreArray(labelsloc,&labels_array); CHKERRQ(info);

  info = VecDestroy(&fx); CHKERRQ(info);
  info = VecDestroy(&fxloc); CHKERRQ(info);
  info = VecDestroy(&labelsloc); CHKERRQ(info);

  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "ScalarLossGradBForce"
// ScalarLossGradBForce - Evaluates a scalar loss and its gradient. 

// Input Parameters:
// w       - current weight vector
// ctx     - optional user-defined context

// Output Parameters:
// G - vector containing the newly evaluated gradient
// f - function value

PetscErrorCode ScalarLossGradBForce(TaoSolver tao, Vec w, double *f, Vec G, void *ctx)
{  
  AppCtx *user=(AppCtx *) ctx;  
  PetscErrorCode info;
  PetscScalar reg=0.0;
  PetscScalar *fx_array, *labels_array, *grad_array;
  
  user->num_fn_eval++;

  if (user->rank == 0)
    user->objgrad_timer.start();

  info = VecDot(w,w,&reg); CHKERRQ(info);
  reg *= user->lambda/2.0;
  
  if (user->rank == 0)
    user->matvec_timer.start();
  
  info = MatMult(user->data,w,user->fx); CHKERRQ(info);

  if (user->rank == 0)
    user->matvec_timer.stop();

  // All nodes send fx to the master node
  info = VecScatterBegin(user->scatter,user->fx,user->fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);
  info = VecScatterEnd(user->scatter,user->fx,user->fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);
    
  // Master node computes the function and gradient coefficients 
  double partloss = 0.0;
  if (user->rank == 0){
    info = VecGetArray(user->fxloc,&fx_array); CHKERRQ(info);
    info = VecGetArray(user->labelsloc,&labels_array); CHKERRQ(info);
    info = VecGetArray(user->gradloc,&grad_array);  CHKERRQ(info);
    
    loss_grad(fx_array,labels_array,grad_array,user->m,&partloss,user);
    
    info = VecRestoreArray(user->fxloc,&fx_array); CHKERRQ(info);
    info = VecRestoreArray(user->labelsloc,&labels_array); CHKERRQ(info);
    info = VecRestoreArray(user->gradloc, &grad_array);  CHKERRQ(info);
  }
  
  MPI_Barrier(PETSC_COMM_WORLD);
  *f = 0.0;
  MPI_Allreduce(&partloss,f,1,MPI_DOUBLE,MPI_SUM,PETSC_COMM_WORLD);
  *f += reg;

  // All nodes receive gradient coefficients   
  info = VecScatterBegin(user->scatter,user->gradloc,user->grad,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(info);
  info = VecScatterEnd(user->scatter,user->gradloc,user->grad,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(info);
  
  // Now everybody can compute the gradient
  if (user->rank == 0)
    user->matvec_timer.start();

  info = MatMultTranspose(user->data, user->grad, G); CHKERRQ(info);

  if (user->rank == 0)
    user->matvec_timer.stop();


  info = VecAXPY(G, user->lambda, w); CHKERRQ(info);
  
  if (user->rank == 0)
    user->objgrad_timer.stop();  
  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "FinalGradLineSearch"
// post-process at the last stage of line search
// Need to return the exact gradient
PetscErrorCode FinalGradLineSearch(TaoLineSearch ls, Vec w, Vec G, void *ctx)
{
  AppCtx *user = (AppCtx *) ctx;  
  PetscErrorCode info;

  user->objgrad_timer.start();

  info = MatMultTranspose(user->data, user->grad, G); CHKERRQ(info);
  info = VecAXPY(G, user->lambda, w);  CHKERRQ(info);

  user->new_round_ls = PETSC_TRUE;
  user->objgrad_timer.stop();
  return 0;
}



#undef __FUNCT__
#define __FUNCT__ "LineSearchObjAndGTS"
PetscErrorCode LineSearchObjAndGTS(TaoLineSearch ls, Vec w, Vec dir, PetscReal *f, PetscReal *gts, void *ctx)
{
  AppCtx *user = (AppCtx *) ctx;   
  PetscErrorCode info;
  PetscReal step;
  PetscScalar *fx_array, *labels_array, *grad_array;
  double partloss;
  Vec lsStart;

  if (user->rank == 0)
    user->objgrad_timer.start();

  info=TaoLineSearchGetStepLength(ls, &step); CHKERRQ(info);

  if (user->new_round_ls)   {
    info = TaoLineSearchGetStartingVector(ls, &lsStart); CHKERRQ(info);

    info = MatMult(user->data, lsStart, user->XMultW); CHKERRQ(info);
    info = MatMult(user->data, dir, user->XMultDir); CHKERRQ(info);

    info = VecDot(lsStart, lsStart, &user->wsq); CHKERRQ(info);  
    info = VecDot(dir, lsStart, &user->dTimesw); CHKERRQ(info);    
    info = VecDot(dir, dir, &user->dsq); CHKERRQ(info);
    
    user->new_round_ls = PETSC_FALSE;
  }

  // First recover the X * (w + stepsize * dir)
  user->num_fn_eval++;

  info=VecCopy(user->XMultW, user->fx); CHKERRQ(info);
  info=VecAXPY(user->fx, step, user->XMultDir); CHKERRQ(info);

  // All nodes send fx to the master node
  info=VecScatterBegin(user->scatter,user->fx,user->fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);
  info=VecScatterEnd(user->scatter,user->fx,user->fxloc,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(info);

  // Master node computes the function and gradient
  if(user->rank==0)
  {
    info=VecGetArray(user->fxloc,&fx_array); CHKERRQ(info);
    info=VecGetArray(user->labelsloc,&labels_array); CHKERRQ(info);
    info=VecGetArray(user->gradloc, &grad_array);  CHKERRQ(info);

    loss_grad(fx_array,labels_array,grad_array,user->m,&partloss,user);

    info=VecRestoreArray(user->fxloc,&fx_array); CHKERRQ(info);
    info=VecRestoreArray(user->labelsloc,&labels_array); CHKERRQ(info);
    info=VecRestoreArray(user->gradloc, &grad_array);  CHKERRQ(info);
  }
  else
    partloss=0.0;

  MPI_Barrier(PETSC_COMM_WORLD);
  *f=0.0;
  MPI_Allreduce(&partloss,f,1,MPI_DOUBLE,MPI_SUM,PETSC_COMM_WORLD);

  // All nodes receive gradient coefficients   
  info=VecScatterBegin(user->scatter,user->gradloc,user->grad,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(info);
  info=VecScatterEnd(user->scatter,user->gradloc,user->grad,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(info);

  *f += 0.5 * user->lambda * (user->wsq + step*step*user->dsq + 2.0*step*user->dTimesw); 
  // printf("Final f = %g\n", *f);
  
  VecDot(user->XMultDir, user->grad, gts);    CHKERRQ(info);
  *gts += user->lambda * (user->dTimesw + step * user->dsq);

  if (user->rank == 0)
    user->objgrad_timer.stop();
  return 0;
}

