/* Copyright (c) 2011, S V N Vishwanathan
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Authors: S V N Vishwanathan (vishy@stat.purdue.edu)
 *          Xinhua Zhang (xinhua.zhang.cs@gmail.com)
 *
*/

#include "hingeloss.hpp"

namespace hingeloss{
  
  // loss_grad - Evaluates smoothed hinge loss and its gradient. 
  
  // Input Parameters:
  // fx_array - function prediction
  // labels_array - labels
  // n - size of arrays
  // user - user defined application context 

  // Output parameters:
  // grad_array - coefficients of the gradient 
  // loss function value 

  PetscErrorCode loss_grad(PetscScalar *fx_array, 
                           PetscScalar *labels_array, 
                           PetscScalar *grad_array, 
                           PetscInt    n,
                           PetscScalar *loss,
                           AppCtx      *user)
  {
    *loss=0.0;
    for (PetscInt i=0; i<n; i++){
      double u=fx_array[i]*labels_array[i];
      if (1.0<u){
        grad_array[i]=0.0;
      } else if (1-u>2*user->eps){
        *loss+=(1-u-user->eps);
        grad_array[i]=-1.0*labels_array[i]/user->m;
      } else {
        double offset=(1-u)/(2*user->eps);
        *loss+=(1-u)*offset/2; 
        grad_array[i]=-offset*labels_array[i]/user->m;
      }
    }
    *loss/=user->m;
    return 0;
  }
  

  // orig_loss - Evaluates the original nonsmooth hinge loss. 

  // Input Parameters:
  // fx_array - function prediction
  // labels_array - labels
  // n - size of arrays
  // user - user defined application context 

  // Output parameters:
  // loss function value 

  PetscErrorCode orig_loss(PetscScalar *fx_array, 
                           PetscScalar *labels_array, 
                           PetscInt    n, 
                           PetscScalar *loss,
                           AppCtx      *user)
  {
    *loss=0.0;
    for (PetscInt i=0; i<n; i++){
      double u=fx_array[i]*labels_array[i];
      if (1.0>=u)
        *loss+=(1.0-u);
    }
    *loss/=user->m;
    return 0;
  }

  PetscErrorCode init(AppCtx *user){ 
    return 0;
  }

  PetscErrorCode finalize(void){
    return 0;
  }
}
