/* Copyright (c) 2011, S V N Vishwanathan
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Authors: S V N Vishwanathan (vishy@stat.purdue.edu)
 *          Xinhua Zhang (xinhua.zhang.cs@gmail.com)
 *
*/

#include "loadfile.hpp"

char* readline(char* line,PetscInt* maxlen,FILE *input){

  PetscInt read_len;

  // First try to read a line from the input file stream. maxlen
  // represents our best guess for the maximum length of the line. If
  // the fgets call returns a NULL then it failed. So we also return a
  // NULL. 
  if (fgets(line,*maxlen,input)==NULL)
    return NULL;


  // If strrchr returns a NULL it means the call to fgets terminated
  // because the line contains more than maxlen characters.
  while (strrchr(line,'\n')==NULL){
    // Double the buffer size until it is large enough to hold the
    // entire input
    char* line_temp=0;
    PetscErrorCode info;
    info=PetscMalloc(*maxlen*2*sizeof(char),&line_temp);
    PetscMemcpy(line_temp,line,*maxlen*sizeof(char));
    PetscFree(line);
    line=line_temp;
    *maxlen*=2;

    read_len=(PetscInt) strlen(line);
    // Read the rest of the line into the reallocated buffer 
    if (fgets(line+read_len,*maxlen-read_len,input)==NULL)
      break;
  }

  return line;
}

PetscErrorCode parse_file(FILE* input,AppCtx_file* user){

  PetscErrorCode info;
  user->maxlen=1024;
  char* line=0;
  info=PetscMalloc(user->maxlen*sizeof(char),&line);CHKERRQ(info);
  PetscInt m_guess=128;
  user->nnz_array=0;
  info=PetscMalloc(m_guess*sizeof(PetscInt),&user->nnz_array);CHKERRQ(info);

  user->dim=-1;
  user->m=0;
  user->maxnnz=0;

  while ((line=readline(line,&user->maxlen,input))!=NULL){

    // Read and ignore the label
    strtok(line," \t");
    PetscInt nnz=0;
    // Now process the rest of the line
    while (1){
      char* pidx=strtok(NULL,":");
      char* pval=strtok(NULL," \t");

      if (pval==NULL)
        break;
      PetscInt idx=(PetscInt) strtol(pidx,NULL,10);
      if (idx > user->dim)
        user->dim=idx;
      nnz++;
    }

    if (user->m >= m_guess){
      PetscInt* nnz_array_temp=0;
      info=PetscMalloc(m_guess*2*sizeof(PetscInt),&nnz_array_temp); CHKERRQ(info);
      info=PetscMemcpy(nnz_array_temp,user->nnz_array,m_guess*sizeof(PetscInt)); CHKERRQ(info);
      info=PetscFree(user->nnz_array); CHKERRQ(info);
      user->nnz_array=nnz_array_temp;
      m_guess *= 2;
    }
    user->nnz_array[user->m]=nnz;

    // Add one extra dimension if needed 
    if (user->flg)
      user->nnz_array[user->m]++;

    // Remember maximum number of nonzeros in any row 
    if(user->nnz_array[user->m]>user->maxnnz)
      user->maxnnz=user->nnz_array[user->m];

    user->m++;
  }

  // adjust for last index
  // Add one extra dimension if needed 
  if (user->flg){
    user->dim+=2;
  } else {
    user->dim+=1;
  }
  // adjust to final size of the array
  PetscInt* nnz_array_temp;
  info=PetscMalloc(user->m*sizeof(PetscInt),&nnz_array_temp); CHKERRQ(info);
  PetscMemcpy(nnz_array_temp,user->nnz_array,user->m*sizeof(PetscInt));
  PetscFree(user->nnz_array);
  user->nnz_array=nnz_array_temp;

  info=PetscFree(line);CHKERRQ(info);
  return 0;
}

void parse_line(char* line,PetscInt* idxs,PetscScalar* vals,PetscScalar* label,int m,AppCtx_file* user){

  // Read the label
  char *plabel=strtok(line," \t");
  *label=strtol(plabel,NULL,10);
  // process the line
  PetscInt nnz=0;
  while (1){
    char* pidx=strtok(NULL,":");
    char* pval=strtok(NULL," \t");
    if (pval==NULL)
      break;
    vals[nnz]=strtod(pval,NULL);
    idxs[nnz]=(PetscInt) strtol(pidx,NULL,10);
    nnz++;
  }
  // Add an extra dimension and set it to one
  if (user->flg){
    vals[nnz]=1.0;
    idxs[nnz]=user->dim-1;
    nnz++;
  }
  assert(user->nnz_array[m]==nnz);
  return;

}

PetscErrorCode assemble_matrix(FILE* input,PetscInt begin,PetscInt end,PetscInt locm,AppCtx_file* user){
  PetscErrorCode info;
  char* line=0;
  info=PetscMalloc(user->maxlen*sizeof(char),&line);CHKERRQ(info);
  PetscInt m=0;
  PetscInt ii=0;

  PetscScalar* vals=0;
  info=PetscMalloc(user->maxnnz*sizeof(PetscScalar),&vals);CHKERRQ(info);
  PetscInt* idxs=0;
  info=PetscMalloc(user->maxnnz*sizeof(PetscInt),&idxs);CHKERRQ(info);

  while ((line=readline(line,&user->maxlen,input)) != NULL){
    //skip the lines for which this processor is not responsible
    if (m<begin){
      m++;
      continue;
    } else {
      if (m>=end)
        break;
    }
    PetscScalar label=0;
    PetscInt one=1;
    parse_line(line,idxs,vals,&label,m,user);

    info=VecSetValues(*user->labels,one,&m,&label,ADD_VALUES);CHKERRQ(info);
    info=MatSetValues(*user->data,one,&m,user->nnz_array[m],idxs,vals,INSERT_VALUES);CHKERRQ(info);
    m++;
    ii++;
  }
  assert(ii=locm);

  info=PetscFree(vals);CHKERRQ(info);
  info=PetscFree(idxs);CHKERRQ(info);
  info=PetscFree(line);CHKERRQ(info);

  info=VecAssemblyBegin(*user->labels);CHKERRQ(info);
  info=VecAssemblyEnd(*user->labels);CHKERRQ(info);

  info=MatAssemblyBegin(*user->data,MAT_FINAL_ASSEMBLY);CHKERRQ(info);
  info=MatAssemblyEnd(*user->data,MAT_FINAL_ASSEMBLY);CHKERRQ(info);
  return 0;
}

PetscErrorCode fill_arrays(FILE* input,AppCtx_file* user){

  PetscErrorCode info;

  // Allocate space for the labels
  info=VecCreate(PETSC_COMM_WORLD,user->labels);CHKERRQ(info);
  info=VecSetSizes(*(user->labels),PETSC_DECIDE,user->m);CHKERRQ(info);
  info=VecSetFromOptions(*(user->labels));CHKERRQ(info);
  info=PetscObjectSetName((PetscObject) *(user->labels),"Labels");CHKERRQ(info);

  // Allocate space for the data 
  info=MatCreate(PETSC_COMM_WORLD,user->data);CHKERRQ(info);
  info=MatSetSizes(*(user->data),PETSC_DECIDE,PETSC_DECIDE,user->m,user->dim);CHKERRQ(info);
  info=MatSetFromOptions(*(user->data));CHKERRQ(info);
  info=MatSetType(*(user->data),MATSEQAIJ);CHKERRQ(info);
  PetscInt zero=0;
  info=MatSeqAIJSetPreallocation(*(user->data),zero,user->nnz_array);CHKERRQ(info);
  info=PetscObjectSetName((PetscObject) *(user->data),"Data");CHKERRQ(info);

  assemble_matrix(input,0,user->m,user->m,user);
  return 0;
}


PetscErrorCode reparse_file(FILE* input,PetscInt* diag_nnz, PetscInt* offdiag_nnz,AppCtx_file* user){

  PetscErrorCode info;
  char* line=0;
  info=PetscMalloc(user->maxlen*sizeof(char),&line);

  PetscInt begin,end;
  info=MatGetOwnershipRange(*(user->data),&begin,&end);CHKERRQ(info);
  PetscInt locm=end-begin;

  PetscInt cbegin,cend;
  MatGetOwnershipRangeColumn(*(user->data),&cbegin,&cend);CHKERRQ(info);

  PetscInt m=0;
  PetscInt ii=0;

  for (PetscInt i=0; i<locm; i++){
    diag_nnz[i]=0;
    offdiag_nnz[i]=0;
  }

  while ((line=readline(line,&user->maxlen,input))!=NULL){
    //skip the lines for which this processor is not responsible
    if (m<begin){
      m++;
      continue;
    } else {
      if (m>=end)
        break;
    }

    // Read and ignore the label
    strtok(line," \t");
    // Now process the rest of the line
    while (1){
      char* pidx=strtok(NULL,":");
      char* pval=strtok(NULL," \t");

      if (pval==NULL)
        break;
      PetscInt idx=(PetscInt) strtol(pidx,NULL,10);
      if (idx>=cbegin && idx<cend)
        diag_nnz[ii]++;
    }

    if (user->flg && user->dim-1>=cbegin && user->dim-1<cend)
      diag_nnz[ii]++;

    offdiag_nnz[ii]=user->nnz_array[m]-diag_nnz[ii];
    m++;
    ii++;
  }
  info=PetscFree(line);CHKERRQ(info);

  assert(ii==locm);
  return 0;
}


PetscErrorCode fill_arrays_parallel(FILE* input,AppCtx_file* user){

  PetscErrorCode info; 
  // Create labels vector 
  info=VecCreate(PETSC_COMM_WORLD,user->labels);CHKERRQ(info);
  info=VecSetSizes(*(user->labels),PETSC_DECIDE,user->m);CHKERRQ(info);
  info=VecSetFromOptions(*(user->labels));CHKERRQ(info);
  info=VecSetType(*(user->labels),VECMPI);
  info=PetscObjectSetName((PetscObject) *(user->labels),"Labels");CHKERRQ(info);

  // Create data matrix 
  info=MatCreate(PETSC_COMM_WORLD,user->data);CHKERRQ(info);
  info=MatSetSizes(*(user->data),PETSC_DECIDE,PETSC_DECIDE,user->m,user->dim);CHKERRQ(info);
  info=MatSetFromOptions(*(user->data));CHKERRQ(info);
  info=MatSetType(*(user->data),MATMPIAIJ);
  info=PetscObjectSetName((PetscObject) *(user->data),"Data");CHKERRQ(info);

  // Allocate space for the data 
  PetscInt begin,end;
  info=MatGetOwnershipRange(*(user->data),&begin,&end);CHKERRQ(info);
  PetscInt locm=end-begin;

  PetscInt diag_nnz[locm];
  PetscInt offdiag_nnz[locm];

  reparse_file(input,diag_nnz,offdiag_nnz,user);
  rewind(input);
  PetscInt zero=0;
  info=MatMPIAIJSetPreallocation(*(user->data),zero,diag_nnz,zero,offdiag_nnz);CHKERRQ(info);

  assemble_matrix(input,begin,end,locm,user);

  return 0;
}






