/* Copyright (c) 2011, S V N Vishwanathan
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Authors: Xinhua Zhang (xinhua.zhang.cs@gmail.com)
 *          S V N Vishwanathan (vishy@stat.purdue.edu)
 *          Ankan Saha (ankans@gmail.com)
 *          
*/

#include "prbeploss.hpp"
#include <algorithm>

namespace prbep{  

  PetscInt *orig_pidx, *orig_nidx;
  PetscInt *pidx, *nidx;
  PetscInt i_pos, i_neg, s_pos, s_neg;
  PetscInt m_pos, m_neg, two_m_pos, two_m_neg;
  PetscScalar *pos_list, *neg_list, nmu, delta, b_pos, b_neg;
  PetscInt verbose;
  
  indirect_greater_than<PetscScalar> *igt_pos, *igt_neg;


  //  We don't change the physical storage of the pos_list and neg_list
  //  We only sort the index, and put the resulting index in pidx[]
  void sort_lists(AppCtx *user)
  {
    PetscInt i;

    for (i = 0; i < 2 * m_pos; i ++)
      pidx[i] = i;
    for (i = 0; i < 2 * m_neg; i ++)
      nidx[i] = i;

    std::sort(pidx, pidx+m_pos*2, *igt_pos);
    std::sort(nidx, nidx+m_neg*2, *igt_neg);
  }

  PetscScalar compute_g(PetscScalar x, PetscScalar label)
  {
    PetscInt *i, *s, i_max;
    PetscScalar *fx, *b;
    PetscInt *idx;

    if (label > 0.0)    {
      i = &i_pos; s = &s_pos; b = &b_pos;  fx = pos_list;  idx = pidx;  i_max = 2*m_pos-1;
    }
    else    {
      i = &i_neg; s = &s_neg; b = &b_neg;  fx = neg_list;  idx = nidx;  i_max = 2*m_neg-1;
    }

    while (1)    {
      if (fx[idx[*i]] >= x)  {
        if (*i == i_max)  return *b / nmu;
        else if (fx[idx[*i + 1]] < x)  return (*b + *s * (x - fx[idx[*i]])) / nmu;
        else  {
          *b += (fx[idx[*i+1]] - fx[idx[*i]]) * (*s);
          (*i)++;
          if (idx[*i] % 2 == 1)  (*s)--;  else  (*s)++;   // tail: --, head: ++
        }
      }
      else      {
        if (*i == 0)  return 0.0;
        if (idx[*i] % 2 == 1)   (*s)++; else (*s)--;  // tail: ++, head --
        (*i)--;
        *b -= (fx[idx[*i+1]] - fx[idx[*i]]) * (*s);
      }
    }

    return 0.0;
  }

  void init_compute_g()
  {
    i_pos = 0;  s_pos = 1;  b_pos = 0;
    i_neg = 2 * m_neg - 1;  s_neg = 0;   b_neg = -m_neg * nmu;
  }


  // loss_grad - Evaluates smoothed hinge loss and its gradient. 

  // Input Parameters:
  // fx_array - function prediction
  // labels_array - labels
  // n - size of arrays
  // user - user defined application context 

  // Output parameters:
  // grad_array - coefficients of the gradient 
  // loss function value 

  PetscErrorCode loss_grad(PetscScalar *fx_array, 
                           PetscScalar *labels_array, 
                           PetscScalar *grad_array, 
                           PetscInt    n, 
                           PetscScalar *loss,
                           AppCtx      *user)
  {
    PetscInt    i, yidx, pre_yidx;
    PetscScalar theta_pos, theta_neg, g_pos=0, g_neg;
    PetscScalar dtemp=0, c, d, PINF, NINF;
    PetscScalar a_pos, b_pos, a_neg, b_neg;
    PetscScalar slope_pos_num, slope_pos_den, inter_pos, slope_neg_num, slope_neg_den, inter_neg;
    PetscInt idx;

    // Populate the lists of pos/neg values
    for (i = 0; i < user->m_pos; i ++)    {
      pos_list[i<<1] = fx_array[orig_pidx[i]] * (-2.0);
      pos_list[1 | i<<1] = pos_list[i<<1] - nmu;
    }

    for (i = 0; i < user->m_neg; i ++)    {
      neg_list[i<<1] = fx_array[orig_nidx[i]] * 2.0;
      neg_list[1 | i<<1] = neg_list[i<<1] - nmu;
    }

    sort_lists(user);
    init_compute_g();

    PINF = fabs(pos_list[pidx[two_m_pos-1]]) + fabs(neg_list[nidx[two_m_neg-1]]) 
            + fabs(pos_list[pidx[0]]) + fabs(neg_list[nidx[0]]) + 2 * delta;
    NINF = -PINF;

    // Stage One
    for (i = 0; i < two_m_pos; i ++)
    {
      dtemp = pos_list[pidx[i]];
      g_pos = compute_g(dtemp, 1.0);
      g_neg = compute_g(-delta - dtemp, -1.0);
      if (g_pos <= g_neg)
        break;
    }

    if(i == two_m_pos)    {
      a_pos = NINF;  b_pos = pos_list[pidx[two_m_pos-1]];  
      c = -delta - b_pos;   d = PINF;   yidx = 0;
      slope_pos_num = 0.0;  slope_pos_den = 1.0;  inter_pos = -m_pos;
    }
    else if (i == 0)    {
      a_pos = dtemp;  b_pos = PINF;
      c = NINF;  d = -delta - dtemp;    yidx = i_neg;
      slope_pos_num = 0.0;  slope_pos_den = 1.0;  inter_pos = 0.0;
    }
    else    {
      a_pos = dtemp;  b_pos = pos_list[pidx[i-1]];
      slope_pos_num = g_pos - compute_g(b_pos, 1);
      slope_pos_den = a_pos - b_pos;  
      inter_pos = g_pos;
      c = -delta - b_pos;   d = -delta - a_pos;   yidx = i_neg;
    }

    // Stage Two

    while(yidx < two_m_neg && neg_list[nidx[yidx]] > d)
      yidx ++;

    pre_yidx = yidx;

    while(1)    {
      if (yidx == two_m_neg || neg_list[nidx[yidx]] < c)  {
        a_neg = c;
        b_neg = pre_yidx != yidx ? neg_list[nidx[yidx-1]] : d;
        inter_neg = compute_g(c, -1.0);
        break;
      }
      
      dtemp = neg_list[nidx[yidx]];      
      g_pos = compute_g(-delta - dtemp, 1.0);
      g_neg = compute_g(dtemp, -1.0);

      if (g_pos >= g_neg)  {
        a_neg = dtemp;
        b_neg = pre_yidx != yidx ? neg_list[nidx[yidx-1]] : d; 
        inter_neg = g_neg;
        break;
      }

      yidx ++;
    }

    // Solve linear system of (theta_pos, theta_neg)
    if (fabs(a_neg - b_neg) < 1e-8)    {
      theta_neg = a_neg;    theta_pos = -delta - theta_neg;
    }
    else    {
      slope_neg_num = inter_neg - compute_g(b_neg, -1.0);
      slope_neg_den = a_neg - b_neg;

      theta_pos = (slope_pos_num * slope_neg_den * a_pos
                  - slope_neg_num * slope_pos_den * a_neg
                  + (inter_neg - inter_pos) * slope_pos_den * slope_neg_den
                  - slope_neg_num * slope_pos_den * delta) 
                  / (slope_pos_num * slope_neg_den + slope_neg_num * slope_pos_den);
      theta_neg = -delta - theta_pos;      
    }

    // assert(fabs(compute_g(theta_pos, 1) - compute_g(theta_neg, -1)) < 1e-5);
    
    // After getting theta^+ and theta^-, compute the gradient coefficient
    
    *loss = 0.0;    c = -2.0 / user->m;     d = 2.0 * nmu;
    for (i = two_m_pos - 1; i >= 0; i--)    {
      while(pidx[i] & 1)   // not head
        i--;
      idx = pidx[i];      
      dtemp = pos_list[idx] - theta_pos;
      if (dtemp < 0.0)  dtemp = 0.0;
      else if (dtemp > nmu)   {
        *loss += dtemp - nmu / 2.0;
        dtemp = 1.0;
      }
      else   {
        *loss += dtemp * dtemp / d;
        dtemp /= nmu;
      }
      grad_array[orig_pidx[idx / 2]] = dtemp * c;
    }
    
    c = 2.0 / user->m;
    for (i = two_m_neg - 1; i >= 0; i--)    {
      while(nidx[i] & 1)   // not head
        i--;
      idx = nidx[i];      
      dtemp = neg_list[idx] - theta_neg;
      if (dtemp < 0.0)  dtemp = 0.0;
      else if (dtemp > nmu)  {
        *loss += dtemp - nmu / 2.0;
        dtemp = 1.0;
      }
      else   {
        *loss += dtemp * dtemp / d;
        dtemp /= nmu;
      }
      grad_array[orig_nidx[idx / 2]] = dtemp * c;
    }

    *loss /= user->m;
    return 0;
  }

  // orig_loss - Evaluates the original nonsmooth hinge loss. 

  // Input Parameters:
  // fx_array - function prediction
  // labels_array - labels
  // n - size of arrays
  // user - user defined application context 

  // Output parameters:
  // loss function value 

  PetscErrorCode orig_loss(PetscScalar *fx_array, 
                           PetscScalar *labels_array, 
                           PetscInt    n, 
                           PetscScalar *loss,
                           AppCtx      *user)
  {
    PetscScalar dtemp;
    PetscInt i, p, q, minm;

    // Populate the lists of pos/neg values
    for (i = 0; i < user->m_pos; i ++)    {
      pos_list[2*i] = fx_array[orig_pidx[i]] * (-2.0);
      pos_list[2*i+1] = pos_list[2*i] - nmu;
    }

    for (i = 0; i < user->m_neg; i ++)    {
      neg_list[2*i] = fx_array[orig_nidx[i]] * 2.0;
      neg_list[2*i+1] = neg_list[2*i] - nmu;
    }

    sort_lists(user);

    *loss = 0.0;
    p = q = 0;
    minm = std::min(user->m_pos, user->m_neg);

    for (i = 0; i < minm; i ++)    {
      dtemp = delta;
      while(pidx[p] & 1)  // tail
        p++;
      dtemp += pos_list[pidx[p++]];

      while(nidx[q] & 1)  // tail
        q++;
      dtemp += neg_list[nidx[q++]];

      if (dtemp > 0.0)
        *loss += dtemp;
      else
        break;
    }

    *loss /= user->m;
    return 0;
  }
  
  
  PetscErrorCode init(AppCtx *user){

    PetscErrorCode info;
    PetscScalar *labels_array;

    pidx = new PetscInt [2*user->m_pos];
    nidx = new PetscInt [2*user->m_neg];
    info = PetscMalloc(user->m_pos*sizeof(PetscInt), &orig_pidx); CHKERRQ(info);
    info = PetscMalloc(user->m_neg*sizeof(PetscInt), &orig_nidx); CHKERRQ(info);
    info = PetscMalloc(2*user->m_pos*sizeof(PetscScalar), &pos_list); CHKERRQ(info);
    info = PetscMalloc(2*user->m_neg*sizeof(PetscScalar), &neg_list); CHKERRQ(info);

    igt_pos = new indirect_greater_than<PetscScalar>(pos_list);     // in descending order
    igt_neg = new indirect_greater_than<PetscScalar>(neg_list);     // in descending order

    info=VecGetArray(user->labelsloc, &labels_array);  CHKERRQ(info);

    PetscInt p(0), n(0);
    for(PetscInt i=0; i<user->m; i++)    {
      if(labels_array[i] > 0.0)
        orig_pidx[p++] = i;
      else
        orig_nidx[n++] = i;
    }

    info=VecRestoreArray(user->labelsloc,&labels_array);  CHKERRQ(info);

    m_pos = user->m_pos;
    m_neg = user->m_neg;
    two_m_pos = 2 * m_pos;
    two_m_neg = 2 * m_neg;
    nmu = 2.0 * user->eps;
    delta = user->m / (PetscScalar) m_pos;
    verbose = 0;

    printf("init of prbep finished, nmu = %g, delta = %g, m_pos = %ld, m_neg = %ld\n", 
            nmu, delta, (long int) m_pos, (long int) m_neg);
    return 0;
  }
  
  PetscErrorCode finalize(void){
    PetscErrorCode info;
    info = PetscFree(orig_pidx); CHKERRQ(info);
    info = PetscFree(orig_nidx); CHKERRQ(info);
    info = PetscFree(pos_list); CHKERRQ(info);
    info = PetscFree(neg_list); CHKERRQ(info);

    delete [] pidx;    delete [] nidx;
    delete igt_pos;    delete igt_neg;
    return 0;
  }
}
