/* Copyright (c) 2011, S V N Vishwanathan
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Authors: S V N Vishwanathan (vishy@stat.purdue.edu)
 *          Ankan Saha (ankans@gmail.com)
 *          Xinhua Zhang (xinhua.zhang.cs@gmail.com)
 *
*/



#include "rocscoreloss.hpp"
#include <algorithm>


namespace rocscore{
  
  // Index vectors used in sorting of fx
  PetscInt *idx=0;
  
  // Product of positive and negative data points
  PetscScalar m_prod;

  // Some intermediate storage variables 
  PetscScalar  *f_array=0;
  PetscInt     *orig_idplus=0;
  PetscInt     *orig_idminus=0;
  PetscInt     *idplus=0;
  PetscInt     *idminus=0;
  PetscScalar  *aplus=0;
  PetscScalar  *aminus=0;
  PetscScalar mu;


  // loss_grad - Evaluates smoothed hinge loss and its gradient. 

  // Input Parameters:
  // fx_array - function prediction
  // labels_array - labels
  // n - size of arrays
  // user - user defined application context 

  // Output parameters:
  // grad_array - coefficients of the gradient 
  // loss function value 

  PetscErrorCode loss_grad(PetscScalar *fx_array, 
                           PetscScalar *labels_array, 
                           PetscScalar *grad_array, 
                           PetscInt    n, 
                           PetscScalar *loss,
                           AppCtx      *user)
  {
    PetscInt i;
    PetscScalar *aplus_rec, *aminus_rec, *ptr;
    aplus_rec = aplus;
    aminus_rec = aminus;

    assert(user->m==n);
    
    PetscInt p_idx=0,n_idx=0;
    
    for (i=0; i<user->m; i++){
      if (labels_array[i]>0.0){
        orig_idplus[p_idx]=i;
        idplus[p_idx]=p_idx;
        aplus[p_idx]=(fx_array[i]-0.5)/(2.0*user->eps);
        p_idx++;
      } else{
        orig_idminus[n_idx]=i;
        idminus[n_idx]=n_idx;
        aminus[n_idx]=(fx_array[i]+0.5)/(2.0*user->eps);
        n_idx++;
      }
    }

    indirect_less_than<PetscScalar> iltplus(aplus), iltminus(aminus);    
    std::sort(idplus, idplus+user->m_pos, iltplus);
    ptr = aplus;
    aplus = f_array;    // borrow the space of f_array, since f_array is not used here
    for (i = 0; i < user->m_pos; i ++)
        aplus[i] = ptr[idplus[i]];    

    std::sort(idminus, idminus+user->m_neg, iltminus);
    ptr = aminus;
    aminus = f_array + user->m_pos;
    for (i = 0; i < user->m_neg; i ++)
        aminus[i] = ptr[idminus[i]];    
    
    double obj=0.0;
    *loss=0.0;

    //Calculate gamma_+
    PetscInt k=-1,j=-1;
    PetscScalar cum_sum=0.0,cum_sum_sq=0.0;
    
    for (i=0; i<user->m_pos; i++){
      while ((j<user->m_neg-1) && (aminus[j+1]<aplus[i])){
        j++;
        cum_sum-=aminus[j];
        cum_sum_sq-=aminus[j]*aminus[j];
      }
      
      while ((k<user->m_neg-1) && (aminus[k+1]<=aplus[i]+1.0)){
        k++;
        cum_sum+=aminus[k];
        cum_sum_sq+=aminus[k]*aminus[k];
      }
      
      PetscInt mip = (user->m_neg-1)-k;
      PetscInt mi = k-j;
      
      PetscScalar gamma=((PetscScalar)mip-mi*aplus[i]+cum_sum)/m_prod;  

      *loss+=(PetscScalar)mip + mi*(aplus[i]*aplus[i]) + cum_sum_sq - 2.0*aplus[i]*cum_sum;

      //printf("Loss first %f\n",*loss);
      obj-=aplus[i]*gamma;
      grad_array[orig_idplus[idplus[i]]] = -gamma;
    }
    
    // Calculate gamma_minus
    k = j = -1;
    cum_sum = 0.0;
    for (PetscInt i=0; i<user->m_neg; i++){
      while ((j<user->m_pos-1) && (aplus[j+1]<aminus[i]-1.0))
        cum_sum-=aplus[++j];
      
      while ((k<user->m_pos-1) && (aplus[k+1]<=aminus[i]))
        cum_sum+=aplus[++k];

      PetscInt mip = j+1;
      PetscInt mi = k-j;

      PetscScalar gamma=((PetscScalar)mip + mi*aminus[i] - cum_sum)/m_prod;  
      obj+=aminus[i]*gamma;
      grad_array[orig_idminus[idminus[i]]]=gamma;
    }
    obj*=2.0*user->eps;             
    *loss*=-(mu/2.0);
    *loss+= obj;

    aplus = aplus_rec;
    aminus = aminus_rec;
    return 0;
  }

  // orig_loss - Evaluates the original nonsmooth hinge loss. 

  // Input Parameters:
  // fx_array - function prediction
  // labels_array - labels
  // n - size of arrays
  // user - user defined application context 

  // Output parameters:
  // loss function value 
  
  PetscErrorCode orig_loss(PetscScalar *fx_array, 
                           PetscScalar *labels_array, 
                           PetscInt    n, 
                           PetscScalar *loss,
                           AppCtx      *user)
  {
    PetscInt m=user->m;
    assert(m==n);
    *loss=0.0;
    memcpy(f_array,fx_array,sizeof(PetscScalar)*m);
    for (PetscInt i=0; i<m; i++){
      f_array[i]-=0.5*labels_array[i];
      idx[i]=i;
    }

    indirect_less_than<PetscScalar> ilt(f_array);
    std::sort(idx, idx+user->m, ilt);
    
    PetscInt l_neg=0;
    PetscInt l_pos=0;
    for (PetscInt i=0; i<m; i++){
      if (labels_array[idx[i]]>0.0) {
        *loss+=f_array[idx[i]]*(l_neg-(double)user->m_neg);
        l_pos++;
      } else{
        *loss+=f_array[idx[i]]*l_pos;
        l_neg++;
      }            
    }
    *loss/=m_prod;
    return 0;
  }
  
  
  PetscErrorCode init(AppCtx *user){ 
    PetscErrorCode info;
    info=PetscMalloc(user->m*sizeof(PetscInt), &idx);CHKERRQ(info);
    info=PetscMalloc(user->m*sizeof(PetscScalar), &f_array);CHKERRQ(info);
    info=PetscMalloc(user->m_pos*sizeof(PetscInt),&orig_idplus);CHKERRQ(info);
    info=PetscMalloc(user->m_pos*sizeof(PetscInt),&idplus);CHKERRQ(info);
    info=PetscMalloc(user->m_pos*sizeof(PetscScalar),&aplus);CHKERRQ(info);
    info=PetscMalloc(user->m_neg*sizeof(PetscInt),&orig_idminus);CHKERRQ(info);
    info=PetscMalloc(user->m_neg*sizeof(PetscInt),&idminus);CHKERRQ(info);
    info=PetscMalloc(user->m_neg*sizeof(PetscScalar),&aminus);CHKERRQ(info);
    m_prod = (PetscScalar)(user->m_pos)*(PetscScalar)(user->m_neg);
    
    mu=2.0*user->eps/m_prod;

    printf("Initialization of ROCArea finished, m_pos = %ld, m_neg = %ld\n", 
            (long int) user->m_pos, (long int) user->m_neg);
    return 0;
  }
  
  PetscErrorCode finalize(void){
    PetscErrorCode info;
    info=PetscFree(idx);CHKERRQ(info);
    info=PetscFree(f_array);CHKERRQ(info);
    
    info=PetscFree(orig_idplus);CHKERRQ(info);
    info=PetscFree(idplus);CHKERRQ(info);
    info=PetscFree(aplus);CHKERRQ(info);

    info=PetscFree(orig_idminus);CHKERRQ(info);
    info=PetscFree(idminus);CHKERRQ(info);
    info=PetscFree(aminus);CHKERRQ(info);
    return 0;
  }
}
